/* Output TOPS-20 format symbol table information from the GNU C compiler.
   Copyright (C) 2002 Free Software Foundation, Inc.
   Contributed by Lars Brinkhoff (lars@nocrew.org), funded by XKL, LLC.

This file is part of GCC.

GCC is free software; you can redistribute it and/or modify it under
the terms of the GNU General Public License as published by the Free
Software Foundation; either version 2, or (at your option) any later
version.

GCC is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License
along with GCC; see the file COPYING.  If not, write to the Free
Software Foundation, 59 Temple Place - Suite 330, Boston, MA
02111-1307, USA.  */

#include "config.h"

#ifdef TOPS20_DEBUGGING_INFO

#include "system.h"
#include "tree.h"
#include "rtl.h"
#include "output.h"
#include "expr.h"
#include "tm_p.h"
#include "debug.h"

#define MAYBE_DECL_NAME(DECL) \
  DECL_NAME (DECL) ? IDENTIFIER_POINTER (DECL_NAME (DECL)) : "(no name)"

static void t20_init 		PARAMS ((const char *));
static void t20_finish		PARAMS ((const char *));
static void t20_define	        PARAMS ((unsigned int, const char *));
static void t20_undef	        PARAMS ((unsigned int, const char *));
static void t20_start_source_file
				PARAMS ((unsigned, const char *));
static void t20_end_source_file	PARAMS ((unsigned));
static void t20_begin_block	PARAMS ((unsigned, unsigned));
static void t20_end_block	PARAMS ((unsigned, unsigned));
static bool t20_ignore_block	PARAMS ((tree));
static void t20_source_line	PARAMS ((unsigned int line,
					 const char * filename));
static void t20_begin_prologue	PARAMS ((unsigned int line,
					 const char *file));
static void t20_end_prologue	PARAMS ((unsigned int line));
static void t20_end_epilogue	PARAMS ((void));
static void t20_begin_function	PARAMS ((tree decl));
static void t20_end_function	PARAMS ((unsigned int line));
static void t20_function_decl	PARAMS ((tree decl));
static void t20_global_decl	PARAMS ((tree decl));
static void t20_deferred_inline_function
				PARAMS ((tree decl));
static void t20_outlining_inline_function
				PARAMS ((tree decl));
static void t20_label		PARAMS ((rtx label));

static inline int is_tagged_type
				PARAMS ((tree type));
static tree decl_ultimate_origin
				PARAMS ((tree decl));

static void
t20_init (main_input_filename)
     const char *main_input_filename;
{
  fprintf (asm_out_file, "\t%s DEBUG: init\n", ASM_COMMENT_START);
}

static void
t20_finish (input_filename)
     const char *input_filename ATTRIBUTE_UNUSED;
{
  fprintf (asm_out_file, "\t%s DEBUG: finish\n", ASM_COMMENT_START);
}

/* Called from debug_define in toplev.c.  The `buffer' parameter contains
   the tail part of the directive line, i.e. the part which is past the
   initial whitespace, #, whitespace, directive-name, whitespace part.  */

static void
t20_define (lineno, buffer)
     unsigned lineno ATTRIBUTE_UNUSED;
     const char *buffer ATTRIBUTE_UNUSED;
{
  fprintf (asm_out_file, "\t%s DEBUG: line %u #define %s\n",
	   ASM_COMMENT_START, lineno, buffer);
}

/* Called from debug_undef in toplev.c.  The `buffer' parameter contains
   the tail part of the directive line, i.e. the part which is past the
   initial whitespace, #, whitespace, directive-name, whitespace part.  */

static void
t20_undef (lineno, buffer)
     unsigned lineno ATTRIBUTE_UNUSED;
     const char *buffer ATTRIBUTE_UNUSED;
{
  fprintf (asm_out_file, "\t%s DEBUG: line %u #undef %s\n",
	   ASM_COMMENT_START, lineno, buffer);
}

/* Record the beginning of a new source file.  */

static void
t20_start_source_file (lineno, filename)
     unsigned int lineno;
     const char *filename;
{
  fprintf (asm_out_file, "\t%s DEBUG: start source file, line %u, file %s\n",
	   ASM_COMMENT_START, lineno, filename);
}

/* Record the end of a source file.  */

static void
t20_end_source_file (lineno)
     unsigned int lineno ATTRIBUTE_UNUSED;
{
  fprintf (asm_out_file, "\t%s DEBUG: end source file, line %u\n",
	   ASM_COMMENT_START, lineno);
}

/* Output a marker (i.e. a label) for the beginning of the generated code for
   a lexical block.  */

static void
t20_begin_block (line, blocknum)
     unsigned int line ATTRIBUTE_UNUSED;
     unsigned int blocknum;
{
  function_section (current_function_decl);
  fprintf (asm_out_file, "\t%s DEBUG: begin block, line %u, block %u\n",
	   ASM_COMMENT_START, line, blocknum);
  ASM_OUTPUT_DEBUG_LABEL (asm_out_file, "BB", blocknum);
}

/* Output a marker (i.e. a label) for the end of the generated code for a
   lexical block.  */

static void
t20_end_block (line, blocknum)
     unsigned int line ATTRIBUTE_UNUSED;
     unsigned int blocknum;
{
  function_section (current_function_decl);
  fprintf (asm_out_file, "\t%s DEBUG: end block, line %u, block %u\n",
	   ASM_COMMENT_START, line, blocknum);
  ASM_OUTPUT_DEBUG_LABEL (asm_out_file, "BE", blocknum);
}

/* Return non-zero if the given type node represents a tagged type.  */

static inline int
is_tagged_type (type)
     tree type;
{
  enum tree_code code = TREE_CODE (type);

  return (code == RECORD_TYPE || code == UNION_TYPE
	  || code == QUAL_UNION_TYPE || code == ENUMERAL_TYPE);
}

/* Determine the "ultimate origin" of a decl.  The decl may be an inlined
   instance of an inlined instance of a decl which is local to an inline
   function, so we have to trace all of the way back through the origin chain
   to find out what sort of node actually served as the original seed for the
   given block.  */

static tree
decl_ultimate_origin (decl)
     tree decl;
{
  /* output_inline_function sets DECL_ABSTRACT_ORIGIN for all the
     nodes in the function to point to themselves; ignore that if
     we're trying to output the abstract instance of this function.  */
  if (DECL_ABSTRACT (decl) && DECL_ABSTRACT_ORIGIN (decl) == decl)
    return NULL_TREE;

#ifdef ENABLE_CHECKING
  if (DECL_FROM_INLINE (DECL_ORIGIN (decl)))
    /* Since the DECL_ABSTRACT_ORIGIN for a DECL is supposed to be the
       most distant ancestor, this should never happen.  */
    abort ();
#endif

  return DECL_ABSTRACT_ORIGIN (decl);
}

/* Define a macro which returns non-zero for a TYPE_DECL which was
   implicitly generated for a tagged type.

   Note that unlike the gcc front end (which generates a NULL named
   TYPE_DECL node for each complete tagged type, each array type, and
   each function type node created) the g++ front end generates a
   _named_ TYPE_DECL node for each tagged type node created.
   These TYPE_DECLs have DECL_ARTIFICIAL set, so we know not to
   generate a DW_TAG_typedef DIE for them.  */

#define TYPE_DECL_IS_STUB(decl)				\
  (DECL_NAME (decl) == NULL_TREE			\
   || (DECL_ARTIFICIAL (decl)				\
       && is_tagged_type (TREE_TYPE (decl))		\
       && ((decl == TYPE_STUB_DECL (TREE_TYPE (decl)))	\
	   /* This is necessary for stub decls that	\
	      appear in nested inline functions.  */	\
	   || (DECL_ABSTRACT_ORIGIN (decl) != NULL_TREE	\
	       && (decl_ultimate_origin (decl)		\
		   == TYPE_STUB_DECL (TREE_TYPE (decl)))))))

/* Returns nonzero if it is appropriate not to emit any debugging
   information for BLOCK, because it doesn't contain any instructions.

   Don't allow this for blocks with nested functions or local classes
   as we would end up with orphans, and in the presence of scheduling
   we may end up calling them anyway.  */

static bool
t20_ignore_block (block)
     tree block;
{
  tree decl;

  for (decl = BLOCK_VARS (block); decl; decl = TREE_CHAIN (decl))
    if (TREE_CODE (decl) == FUNCTION_DECL
	|| (TREE_CODE (decl) == TYPE_DECL && TYPE_DECL_IS_STUB (decl)))
      return 0;

  return 1;
}

/* Output a label to mark the beginning of a source code line entry
   and record information relating to this source line.  */

static void
t20_source_line (line, filename)
     unsigned int line;
     const char *filename;
{
  fprintf (asm_out_file, "\t%s DEBUG: source line %u, file %s\n",
	   ASM_COMMENT_START, line, filename);
#if 0
  fprintf (asm_out_file, "label:\n");
  fprintf (asm_out_file, "\t.ENDPS\n");
  fprintf (asm_out_file, "\tPSECT DBGLIN\n");
  fprintf (asm_out_file, "\tEXP label,%u,[ASCIZ /%s/]\n", line, filename);
  fprintf (asm_out_file, "\t.ENDPS\n");
  fprintf (asm_out_file, "\tPSECT <previous>\n");
#endif
}

/* Output a marker (i.e. a label) for the beginning of a function, before
   the prologue.  */

static void
t20_begin_prologue (line, file)
     unsigned int line ATTRIBUTE_UNUSED;
     const char *file ATTRIBUTE_UNUSED;
{
  fprintf (asm_out_file, "\t%s DEBUG: begin prologue, line %u file %s\n",
	   ASM_COMMENT_START, line, file);
}

static void
t20_end_prologue (line)
     unsigned int line;
{
  fprintf (asm_out_file, "\t%s DEBUG: end prologue, line %u\n",
	   ASM_COMMENT_START, line);
}

/* Output a marker (i.e. a label) for the absolute end of the generated code
   for a function definition.  This gets called *after* the epilogue code has
   been generated.  */

static void
t20_end_epilogue ()
{
  fprintf (asm_out_file, "\t%s DEBUG: end prologue\n", ASM_COMMENT_START);
}

static void
t20_begin_function (decl)
     tree decl;
{
  fprintf (asm_out_file, "\t%s DEBUG: begin function \"%s\"\n",
	   ASM_COMMENT_START, MAYBE_DECL_NAME (decl));
}

static void
t20_end_function (line)
     unsigned int line;
{
  fprintf (asm_out_file, "\t%s DEBUG: end function, line %u\n",
	   ASM_COMMENT_START, line);
}

/* Write the debugging output for DECL.  */

static void
t20_function_decl (decl)
     tree decl;
{
  fprintf (asm_out_file, "\t%s DEBUG: ", ASM_COMMENT_START);

  switch (TREE_CODE (decl))
    {
    case ERROR_MARK:
      fprintf (asm_out_file, "error_mark\n");
      return;

    case FUNCTION_DECL:
      fprintf (asm_out_file, "function_decl \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    case VAR_DECL:
      fprintf (asm_out_file, "var_decl \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    case TYPE_DECL:
      fprintf (asm_out_file, "type_decl \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    default:
      fprintf (asm_out_file, "some other decl?\n");
      return;
    }
}

/* Output debug information for global decl DECL.  Called from
   toplev.c after compilation proper has finished.  */

static void
t20_global_decl (decl)
     tree decl;
{
  fprintf (asm_out_file, "\t%s DEBUG: global_decl ", ASM_COMMENT_START);

  switch (TREE_CODE (decl))
    {
    case ERROR_MARK:
      fprintf (asm_out_file, "error_mark\n");
      return;

    case FUNCTION_DECL:
      fprintf (asm_out_file, "function \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    case VAR_DECL:
      fprintf (asm_out_file, "variable \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    case TYPE_DECL:
      fprintf (asm_out_file, "type \"%s\"\n", MAYBE_DECL_NAME (decl));
      return;

    default:
      fprintf (asm_out_file, "something else\n");
      return;
    }
}

static void
t20_deferred_inline_function (decl)
     tree decl;
{
  fprintf (asm_out_file, "\t%s DEBUG: deferred inline function \"%s\"\n",
	   ASM_COMMENT_START, MAYBE_DECL_NAME (decl));
}

static void
t20_outlining_inline_function (decl)
     tree decl;
{
  fprintf (asm_out_file, "\t%s DEBUG: outlining inline function \"%s\"\n",
	   ASM_COMMENT_START, MAYBE_DECL_NAME (decl));
}

static void
t20_label (label)
     rtx label;
{
  fprintf (asm_out_file, "\t%s DEBUG: label \"%s\"\n",
	   ASM_COMMENT_START, LABEL_NAME (label));
}

/* The debug hooks structure.  */

const struct gcc_debug_hooks tops20_debug_hooks =
{
  t20_init,
  t20_finish,
  t20_define,
  t20_undef,
  t20_start_source_file,
  t20_end_source_file,
  t20_begin_block,
  t20_end_block,
  t20_ignore_block,
  t20_source_line,
  t20_begin_prologue,
  t20_end_prologue,
  t20_end_epilogue,
  t20_begin_function,
  t20_end_function,
  t20_function_decl,
  t20_global_decl,
  t20_deferred_inline_function,
  t20_outlining_inline_function,
  t20_label
};

#endif /* TOPS20_DEBUGGING_INFO */
